@@ -21,13 +21,14 @@
 #include "mongoc-client-session-private.h"
 #include "mongoc-counters-private.h"
 #include "mongoc-error.h"
+#include "mongoc-error-private.h"
 #include "mongoc-log.h"
 #include "mongoc-trace-private.h"
 #include "mongoc-read-concern-private.h"
 #include "mongoc-util-private.h"
 #include "mongoc-write-concern-private.h"
 #include "mongoc-read-prefs-private.h"
-
+#include "mongoc-aggregate-private.h"
 
 #undef MONGOC_LOG_DOMAIN
 #define MONGOC_LOG_DOMAIN "cursor"
@@ -163,8 +164,9 @@ _mongoc_set_cursor_ns (mongoc_cursor_t *cursor, const char *ns, uint32_t nslen)
 {
    const char *dot;
 
-   bson_strncpy (cursor->ns, ns, sizeof cursor->ns);
-   cursor->nslen = BSON_MIN (nslen, sizeof cursor->ns);
+   bson_free (cursor->ns);
+   cursor->ns = bson_strndup (ns, nslen);
+   cursor->nslen = nslen;
    dot = strstr (cursor->ns, ".");
 
    if (dot) {
@@ -236,6 +238,7 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
    mongoc_cursor_t *cursor;
    mongoc_topology_description_type_t td_type;
    uint32_t server_id;
+   mongoc_read_concern_t *read_concern_local = NULL;
    bson_error_t validate_err;
    const char *dollar_field;
    bson_iter_t iter;
@@ -247,6 +250,8 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
    cursor = (mongoc_cursor_t *) bson_malloc0 (sizeof *cursor);
    cursor->client = client;
    cursor->state = UNPRIMED;
+   cursor->client_generation = client->generation;
+   cursor->is_aggr_with_write_stage = false;
 
    bson_init (&cursor->opts);
    bson_init (&cursor->error_doc);
@@ -281,6 +286,18 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
          cursor->explicit_session = true;
       }
 
+      if (bson_iter_init_find (&iter, opts, "readConcern")) {
+         read_concern_local =
+            _mongoc_read_concern_new_from_iter (&iter, &cursor->error);
+
+         if (!read_concern_local) {
+            /* invalid read concern */
+            GOTO (finish);
+         }
+
+         read_concern = read_concern_local;
+      }
+
       /* true if there's a valid serverId or no serverId, false on err */
       if (!_mongoc_get_server_id_from_opts (opts,
                                             MONGOC_ERROR_CURSOR,
@@ -294,8 +311,19 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
          (void) mongoc_cursor_set_hint (cursor, server_id);
       }
 
-      bson_copy_to_excluding_noinit (
-         opts, &cursor->opts, "serverId", "sessionId", NULL);
+      bson_copy_to_excluding_noinit (opts,
+                                     &cursor->opts,
+                                     "serverId",
+                                     "sessionId",
+                                     "bypassDocumentValidation",
+                                     NULL);
+
+
+      /* only include bypassDocumentValidation if it's true */
+      if (bson_iter_init_find (&iter, opts, "bypassDocumentValidation") &&
+          bson_iter_as_bool (&iter)) {
+         BSON_APPEND_BOOL (&cursor->opts, "bypassDocumentValidation", true);
+      }
    }
 
    if (_mongoc_client_session_in_txn (cursor->client_session)) {
@@ -356,6 +384,7 @@ _mongoc_cursor_new_with_opts (mongoc_client_t *client,
    (void) _mongoc_read_prefs_validate (cursor->read_prefs, &cursor->error);
 
 finish:
+   mongoc_read_concern_destroy (read_concern_local);
    mongoc_counter_cursors_active_inc ();
 
    RETURN (cursor);
@@ -415,19 +444,19 @@ _translate_query_opt (const char *query_field, const char **cmd_field, int *len)
 
 
 /* set up a new opt bson from older ways of specifying options.
- * slave_ok may be NULL.
+ * secondary_ok may be NULL.
  * error may be NULL.
  */
 void
 _mongoc_cursor_flags_to_opts (mongoc_query_flags_t qflags,
                               bson_t *opts, /* IN/OUT */
-                              bool *slave_ok /* OUT */)
+                              bool *secondary_ok /* OUT */)
 {
    ENTRY;
    BSON_ASSERT (opts);
 
-   if (slave_ok) {
-      *slave_ok = !!(qflags & MONGOC_QUERY_SLAVE_OK);
+   if (secondary_ok) {
+      *secondary_ok = !!(qflags & MONGOC_QUERY_SECONDARY_OK);
    }
 
    if (qflags & MONGOC_QUERY_TAILABLE_CURSOR) {
@@ -560,7 +589,7 @@ done:
 void
 mongoc_cursor_destroy (mongoc_cursor_t *cursor)
 {
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db;
    ENTRY;
 
    if (!cursor) {
@@ -571,23 +600,30 @@ mongoc_cursor_destroy (mongoc_cursor_t *cursor)
       cursor->impl.destroy (&cursor->impl);
    }
 
+   /* Always close the socket for an exhaust cursor, even if the client was
+    * reset with mongoc_client_reset. That prevents further use of that socket.
+    */
    if (cursor->in_exhaust) {
       cursor->client->in_exhaust = false;
       if (cursor->state != DONE) {
-         /* The only way to stop an exhaust cursor is to kill the connection */
-         mongoc_cluster_disconnect_node (
-            &cursor->client->cluster, cursor->server_id, false, NULL);
+         /* The only way to stop an exhaust cursor is to kill the connection
+          */
+         mongoc_cluster_disconnect_node (&cursor->client->cluster,
+                                         cursor->server_id);
       }
-   } else if (cursor->cursor_id) {
-      bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   } else if (cursor->client_generation == cursor->client->generation) {
+      if (cursor->cursor_id) {
+         db = bson_strndup (cursor->ns, cursor->dblen);
 
-      _mongoc_client_kill_cursor (cursor->client,
-                                  cursor->server_id,
-                                  cursor->cursor_id,
-                                  cursor->operation_id,
-                                  db,
-                                  cursor->ns + cursor->dblen + 1,
-                                  cursor->client_session);
+         _mongoc_client_kill_cursor (cursor->client,
+                                     cursor->server_id,
+                                     cursor->cursor_id,
+                                     cursor->operation_id,
+                                     db,
+                                     cursor->ns + cursor->dblen + 1,
+                                     cursor->client_session);
+         bson_free (db);
+      }
    }
 
    if (cursor->client_session && !cursor->explicit_session) {
@@ -600,6 +636,7 @@ mongoc_cursor_destroy (mongoc_cursor_t *cursor)
 
    bson_destroy (&cursor->opts);
    bson_destroy (&cursor->error_doc);
+   bson_free (cursor->ns);
    bson_free (cursor);
 
    mongoc_counter_cursors_active_dec ();
@@ -618,6 +655,8 @@ _mongoc_cursor_fetch_stream (mongoc_cursor_t *cursor)
    ENTRY;
 
    if (cursor->server_id) {
+      /* We already did server selection once before. Reuse the prior
+       * selection to create a new stream on the same server. */
       server_stream =
          mongoc_cluster_stream_for_server (&cursor->client->cluster,
                                            cursor->server_id,
@@ -625,15 +664,26 @@ _mongoc_cursor_fetch_stream (mongoc_cursor_t *cursor)
                                            cursor->client_session,
                                            &reply,
                                            &cursor->error);
+      if (server_stream) {
+         /* Also restore whether primary read preference was forced by server
+          * selection */
+         server_stream->must_use_primary = cursor->must_use_primary;
+      }
    } else {
-      server_stream = mongoc_cluster_stream_for_reads (&cursor->client->cluster,
-                                                       cursor->read_prefs,
-                                                       cursor->client_session,
-                                                       &reply,
-                                                       &cursor->error);
+      server_stream =
+         mongoc_cluster_stream_for_reads (&cursor->client->cluster,
+                                          cursor->read_prefs,
+                                          cursor->client_session,
+                                          &reply,
+                                          cursor->is_aggr_with_write_stage,
+                                          &cursor->error);
 
       if (server_stream) {
+         /* Remember the selected server_id and whether primary read mode was
+          * forced so that we can re-create an equivalent server_stream at a
+          * later time */
          cursor->server_id = server_stream->sd->id;
+         cursor->must_use_primary = server_stream->must_use_primary;
       }
    }
 
@@ -655,7 +705,7 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
 {
    mongoc_client_t *client;
    mongoc_apm_command_started_t event;
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db;
 
    ENTRY;
 
@@ -665,7 +715,7 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
       RETURN (true);
    }
 
-   bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   db = bson_strndup (cursor->ns, cursor->dblen);
 
    mongoc_apm_command_started_init (&event,
                                     cmd,
@@ -675,10 +725,13 @@ _mongoc_cursor_monitor_command (mongoc_cursor_t *cursor,
                                     cursor->operation_id,
                                     &server_stream->sd->host,
                                     server_stream->sd->id,
+                                    &server_stream->sd->service_id,
+                                    NULL,
                                     client->apm_context);
 
    client->apm_callbacks.started (&event);
    mongoc_apm_command_started_cleanup (&event);
+   bson_free (db);
 
    RETURN (true);
 }
@@ -754,6 +807,8 @@ _mongoc_cursor_monitor_succeeded (mongoc_cursor_t *cursor,
                                       cursor->operation_id,
                                       &stream->sd->host,
                                       stream->sd->id,
+                                      &stream->sd->service_id,
+                                      false,
                                       client->apm_context);
 
    client->apm_callbacks.succeeded (&event);
@@ -798,6 +853,8 @@ _mongoc_cursor_monitor_failed (mongoc_cursor_t *cursor,
                                    cursor->operation_id,
                                    &stream->sd->host,
                                    stream->sd->id,
+                                   &stream->sd->service_id,
+                                   false,
                                    client->apm_context);
 
    client->apm_callbacks.failed (&event);
@@ -860,13 +917,13 @@ _mongoc_cursor_opts_to_flags (mongoc_cursor_t *cursor,
       }
    }
 
-   if (cursor->slave_ok) {
-      *flags |= MONGOC_QUERY_SLAVE_OK;
+   if (cursor->secondary_ok) {
+      *flags |= MONGOC_QUERY_SECONDARY_OK;
    } else if (cursor->server_id &&
               (stream->topology_type == MONGOC_TOPOLOGY_RS_WITH_PRIMARY ||
                stream->topology_type == MONGOC_TOPOLOGY_RS_NO_PRIMARY) &&
               stream->sd->type != MONGOC_SERVER_RS_PRIMARY) {
-      *flags |= MONGOC_QUERY_SLAVE_OK;
+      *flags |= MONGOC_QUERY_SECONDARY_OK;
    }
 
    return true;
@@ -876,22 +933,22 @@ bool
 _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
                             const bson_t *command,
                             const bson_t *opts,
-                            bson_t *reply)
+                            bson_t *reply,
+                            bool retry_prohibited)
 {
-   mongoc_cluster_t *cluster;
    mongoc_server_stream_t *server_stream;
    bson_iter_t iter;
    mongoc_cmd_parts_t parts;
    const char *cmd_name;
    bool is_primary;
    mongoc_read_prefs_t *prefs = NULL;
-   char db[MONGOC_NAMESPACE_MAX];
+   char *db = NULL;
    mongoc_session_opt_t *session_opts;
    bool ret = false;
+   bool is_retryable = true;
 
    ENTRY;
 
-   cluster = &cursor->client->cluster;
    mongoc_cmd_parts_init (
       &parts, cursor->client, db, MONGOC_QUERY_NONE, command);
    parts.is_read_command = true;
@@ -920,6 +977,11 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
          _mongoc_bson_init_if_set (reply);
          GOTO (done);
       }
+      if (_mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_EXHAUST)) {
+         MONGOC_WARNING (
+            "exhaust cursors not supported with OP_MSG, using normal "
+            "cursor instead");
+      }
    }
 
    if (parts.assembled.session) {
@@ -953,7 +1015,7 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
       GOTO (done);
    }
 
-   bson_strncpy (db, cursor->ns, cursor->dblen + 1);
+   db = bson_strndup (cursor->ns, cursor->dblen);
    parts.assembled.db_name = db;
 
    if (!_mongoc_cursor_opts_to_flags (
@@ -962,10 +1024,19 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
       GOTO (done);
    }
 
+   /* Exhaust cursors with OP_MSG not yet supported; fallback to normal cursor.
+    * user_query_flags is unused in OP_MSG, so this technically has no effect,
+    * but is done anyways to ensure the query flags match handling of options.
+    */
+   if (parts.user_query_flags & MONGOC_QUERY_EXHAUST) {
+      parts.user_query_flags ^= MONGOC_QUERY_EXHAUST;
+   }
+
    /* we might use mongoc_cursor_set_hint to target a secondary but have no
     * read preference, so the secondary rejects the read. same if we have a
     * direct connection to a secondary (topology type "single"). with
-    * OP_QUERY we handle this by setting slaveOk. here we use $readPreference.
+    * OP_QUERY we handle this by setting secondaryOk. here we use
+    * $readPreference.
     */
    cmd_name = _mongoc_get_command_name (command);
    is_primary =
@@ -973,13 +1044,31 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
 
    if (strcmp (cmd_name, "getMore") != 0 &&
        server_stream->sd->max_wire_version >= WIRE_VERSION_OP_MSG &&
-       is_primary && parts.user_query_flags & MONGOC_QUERY_SLAVE_OK) {
+       is_primary && parts.user_query_flags & MONGOC_QUERY_SECONDARY_OK) {
       parts.read_prefs = prefs =
          mongoc_read_prefs_new (MONGOC_READ_PRIMARY_PREFERRED);
    } else {
       parts.read_prefs = cursor->read_prefs;
    }
 
+   is_retryable = _is_retryable_read (&parts, server_stream);
+   if (!strcmp (cmd_name, "getMore")) {
+      is_retryable = false;
+   }
+   if (!strcmp (cmd_name, "aggregate")) {
+      bson_iter_t pipeline_iter;
+      if (bson_iter_init_find (&pipeline_iter, command, "pipeline") &&
+          BSON_ITER_HOLDS_ARRAY (&pipeline_iter) &&
+          bson_iter_recurse (&pipeline_iter, &pipeline_iter)) {
+         if (_has_write_key (&pipeline_iter)) {
+            is_retryable = false;
+         }
+      }
+   }
+   if (is_retryable && retry_prohibited) {
+      is_retryable = false;
+   }
+
    if (cursor->write_concern &&
        !mongoc_write_concern_is_default (cursor->write_concern) &&
        server_stream->sd->max_wire_version >= WIRE_VERSION_CMD_WRITE_CONCERN) {
@@ -993,8 +1082,40 @@ _mongoc_cursor_run_command (mongoc_cursor_t *cursor,
       GOTO (done);
    }
 
+retry:
    ret = mongoc_cluster_run_command_monitored (
-      cluster, &parts.assembled, reply, &cursor->error);
+      &cursor->client->cluster, &parts.assembled, reply, &cursor->error);
+
+   if (ret) {
+      memset (&cursor->error, 0, sizeof (bson_error_t));
+   }
+
+   if (is_retryable &&
+       _mongoc_read_error_get_type (ret, &cursor->error, reply) ==
+          MONGOC_READ_ERR_RETRY) {
+      is_retryable = false;
+
+      mongoc_server_stream_cleanup (server_stream);
+
+      BSON_ASSERT (!cursor->is_aggr_with_write_stage &&
+                   "Cannot attempt a retry on an aggregate operation that "
+                   "contains write stages");
+      server_stream =
+         mongoc_cluster_stream_for_reads (&cursor->client->cluster,
+                                          cursor->read_prefs,
+                                          cursor->client_session,
+                                          reply,
+                                          /* Not aggregate-with-write */ false,
+                                          &cursor->error);
+
+      if (server_stream &&
+          server_stream->sd->max_wire_version >= WIRE_VERSION_RETRY_READS) {
+         cursor->server_id = server_stream->sd->id;
+         parts.assembled.server_stream = server_stream;
+         bson_destroy (reply);
+         GOTO (retry);
+      }
+   }
 
    if (cursor->error.domain) {
       bson_destroy (&cursor->error_doc);
@@ -1017,6 +1138,7 @@ done:
    mongoc_server_stream_cleanup (server_stream);
    mongoc_cmd_parts_cleanup (&parts);
    mongoc_read_prefs_destroy (prefs);
+   bson_free (db);
 
    return ret;
 }
@@ -1139,6 +1261,14 @@ mongoc_cursor_next (mongoc_cursor_t *cursor, const bson_t **bson)
 
    TRACE ("cursor_id(%" PRId64 ")", cursor->cursor_id);
 
+   if (cursor->client_generation != cursor->client->generation) {
+      bson_set_error (&cursor->error,
+                      MONGOC_ERROR_CURSOR,
+                      MONGOC_ERROR_CURSOR_INVALID_CURSOR,
+                      "Cannot advance cursor after client reset");
+      RETURN (false);
+   }
+
    if (bson) {
       *bson = NULL;
    }
@@ -1224,7 +1354,8 @@ mongoc_cursor_more (mongoc_cursor_t *cursor)
 void
 mongoc_cursor_get_host (mongoc_cursor_t *cursor, mongoc_host_list_t *host)
 {
-   mongoc_server_description_t *description;
+   mongoc_server_description_t const *description;
+   mc_shared_tpld td;
 
    BSON_ASSERT (cursor);
    BSON_ASSERT (host);
@@ -1236,16 +1367,16 @@ mongoc_cursor_get_host (mongoc_cursor_t *cursor, mongoc_host_list_t *host)
       return;
    }
 
-   description = mongoc_topology_server_by_id (
-      cursor->client->topology, cursor->server_id, &cursor->error);
+   td = mc_tpld_take_ref (cursor->client->topology);
+   description = mongoc_topology_description_server_by_id_const (
+      td.ptr, cursor->server_id, &cursor->error);
+   mc_tpld_drop_ref (&td);
    if (!description) {
       return;
    }
 
    *host = description->host;
 
-   mongoc_server_description_destroy (description);
-
    EXIT;
 }
 
@@ -1283,7 +1414,7 @@ mongoc_cursor_clone (const mongoc_cursor_t *cursor)
    bson_copy_to (&cursor->opts, &_clone->opts);
    bson_init (&_clone->error_doc);
 
-   bson_strncpy (_clone->ns, cursor->ns, sizeof _clone->ns);
+   _clone->ns = bson_strdup (cursor->ns);
 
    /* copy the context functions by default. */
    memcpy (&_clone->impl, &cursor->impl, sizeof (cursor->impl));
@@ -1579,7 +1710,8 @@ _mongoc_cursor_response_refresh (mongoc_cursor_t *cursor,
 
    /* server replies to find / aggregate with {cursor: {id: N, firstBatch: []}},
     * to getMore command with {cursor: {id: N, nextBatch: []}}. */
-   if (_mongoc_cursor_run_command (cursor, command, opts, &response->reply) &&
+   if (_mongoc_cursor_run_command (
+          cursor, command, opts, &response->reply, false) &&
        _mongoc_cursor_start_reading_response (cursor, response)) {
       return;
    }
@@ -1592,7 +1724,6 @@ _mongoc_cursor_response_refresh (mongoc_cursor_t *cursor,
    }
 }
 
-
 void
 _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
                                         bson_t *command)
@@ -1601,7 +1732,7 @@ _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
    int collection_len;
    int64_t batch_size;
    bool await_data;
-   int32_t max_await_time_ms;
+   int64_t max_await_time_ms;
 
    ENTRY;
 
@@ -1631,12 +1762,11 @@ _mongoc_cursor_prepare_getmore_command (mongoc_cursor_t *cursor,
    await_data = _mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_TAILABLE) &&
                 _mongoc_cursor_get_opt_bool (cursor, MONGOC_CURSOR_AWAIT_DATA);
 
-
    if (await_data) {
-      max_await_time_ms =
-         (int32_t) mongoc_cursor_get_max_await_time_ms (cursor);
+      max_await_time_ms = _mongoc_cursor_get_opt_int64 (
+         cursor, MONGOC_CURSOR_MAX_AWAIT_TIME_MS, 0);
       if (max_await_time_ms) {
-         bson_append_int32 (command,
+         bson_append_int64 (command,
                             MONGOC_CURSOR_MAX_TIME_MS,
                             MONGOC_CURSOR_MAX_TIME_MS_LEN,
                             max_await_time_ms);
